from django.utils.deprecation import MiddlewareMixin

from .abmain import AB
from .models import Experiment


class ABMiddleware(MiddlewareMixin):
    def process_request(self, request):
        request.ab = AB(request)
        if request.ab.is_active():
            exps = Experiment.objects.filter(cancelled=False)
            for exp in exps:
                if exp.is_active:
                    exp.percentage = exp.get_updated_traffic()
                exp.save()
                if request.ab.is_converted(exp):
                    request.ab.convert(exp)



from __future__ import unicode_literals

from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.core.validators import MaxValueValidator
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _


@python_2_unicode_compatible
class Experiment(models.Model):
    name = models.CharField(max_length=255, unique=True)
    template_name = models.CharField(max_length=255, unique=True,
                                    help_text= ("Example : '/main/signup.html'. Original Template"))
    goal = models.CharField(max_length=255, unique=True,
                                    help_text=_("Example : '/main/signup_complete'. The path on goal Completion"))
    start = models.DateTimeField(blank=False, db_index=True,
                                    help_text=_("This Test starts at the selected Date."))
    end = models.DateTimeField(blank=False, null=True,
                                    help_text=_("This Test ends at the selected Date."))
    percentage = models.PositiveIntegerField(_("Traffic Percentage"), validators=[MaxValueValidator(100), ], default=50,
                                    help_text=_("Percentage of traffic to redirect to Variant B"))
    cancelled = models.BooleanField(_("Cancelled"), default=False)
    created = models.DateTimeField(_("Creation Date and Time"), auto_now_add=True)
    updated = models.DateTimeField(_("Last Update Date and Time"), auto_now=True)

    class Meta:
        verbose_name = _("Experiment")
        verbose_name_plural = _("Experiments")

    def __str__(self):
        return self.name

    def get_experiment_key(self):
        return "ab_exp_key_%s" % self.name

    @property
    def is_active(self):
        return (
            not self.cancelled and
            self.start <= timezone.now() < self.end
        )

    def get_status(self):
        '''
            Returns the Experiment Status based on the Start Date and End Date
        '''
        today = timezone.now()
        if self.start > today or self.end < today:
            active = False
        else:
            active = True
        return active

    def get_updated_traffic(self):
        '''
            Returns the updated network traffic percentage
            for the first Test of the current Experiment
        '''
        test_1, test_2 = self.test_set.all()
        try:
            test_1_ratio = float(test_1.conversions) / test_1.hits
            test_2_ratio = float(test_2.conversions) / test_2.hits
            updated_traffic = int(test_2_ratio*100/(test_1_ratio+test_2_ratio))
            return updated_traffic
        except ZeroDivisionError:
            return 50

    def get_absolute_url(self):
        return reverse(
            "experiment_details", kwargs={
                    "experiment_id": str(self.pk)
                }
        )


@python_2_unicode_compatible
class Test(models.Model):
    experiment = models.ForeignKey(Experiment)
    template_name = models.CharField(max_length=255, unique=True,
                    help_text=_(
                        "Example : '/main/signup1.html'. Template to be Tested"
                            ))
    hits = models.PositiveIntegerField(blank=True, default=0,
                    help_text=_(
                        "# Uniques who have seen this test."
                            ))
    conversions = models.PositiveIntegerField(blank=True, default=0,
                    help_text=_(
                        "# Uniques that have reached the goal from this test."
                            ))

    class Meta:
        verbose_name = _("Test")
        verbose_name_plural = _("Tests")

    def __str__(self):
        return self.template_name

    def save(self, *args, **kwargs):
        tests = Test.objects.filter(experiment=self.experiment).count()
        try:
            current_test = Test.objects.get(id=self.pk)
            super(Test, self).save(*args, **kwargs)
        except:
            pass
            if tests < 2:
                super(Test, self).save(*args, **kwargs)
            else:
                raise ValidationError(
                    "can only create 2 Tests currently for a single Experiment"
                        )



import random

from .models import Experiment, Test


class AB(object):

    def __init__(self, request):
        self.request = request

    def is_active(self):
        return "active" in self.request.session

    def is_converted(self, exp):
        return self._is_experiment_active(exp)\
         and not self._is_experiment_converted(exp) \
         and exp.goal in self.request.path

    def _is_experiment_active(self, exp):
        return exp.get_experiment_key() in self.request.session

    def _is_experiment_converted(self, exp):
        return "converted" \
         in self.request.session[exp.get_experiment_key()]

    def _get_traffic(self, percentage):
        return 0 if random.random() <= percentage else 1

    def _get_test(self, exp):
        tests = exp.test_set.all()
        test_index = self._get_traffic(exp.percentage / 100.0)
        test = tests[test_index]
        return test

    def run(self, template_name):
        try:
            exp = Experiment.objects.get(template_name=template_name)
        except Experiment.DoesNotExist:
            return template_name

        if exp.cancelled:
            return template_name

        key = exp.get_experiment_key()
        if self._is_experiment_active(exp):
            return self.request.session[key]["template"]

        test = self._get_test(exp)
        self.activate(test, key)

        return test.template_name

    def activate(self, test, key):
        test.hits += 1
        test.save()
        self.request.session[key] = {
                "id": test.id,
                "template": test.template_name
            }

        self.request.session["active"] = True

    def convert(self, exp):
        key = exp.get_experiment_key()
        test_id = self.request.session[key]["id"]
        test = Test.objects.get(pk=test_id)
        test.conversions += 1
        test.save()
        self.request.session[key]["converted"] = True
        self.request.session.modified = True
